@@ -4,6 +4,7 @@
 Exercises various features both the Pulsar client and server.
 """
 
+import logging
 import optparse
 import os
 import re
@@ -17,14 +18,17 @@ import traceback
 from collections import namedtuple
 from io import open
 
-from galaxy.tools.deps.dependencies import DependenciesDescription
-from galaxy.tools.deps.requirements import ToolRequirement
+from galaxy.tool_util.deps.dependencies import DependenciesDescription
+from galaxy.tool_util.deps.requirements import ToolRequirement
 from six import binary_type
 
 from pulsar.client import (
     build_client_manager,
     ClientJobDescription,
+    ClientInputs,
+    ClientInput,
     ClientOutputs,
+    CLIENT_INPUT_PATH_TYPES,
     finish_job,
     PulsarOutputs,
     submit_job,
@@ -40,11 +44,15 @@ import sys
 from os import getenv
 from os import listdir
 from os import makedirs
+from os.path import abspath
 from os.path import basename
 from os.path import dirname
 from os.path import exists
 from os.path import join
 
+SCRIPT_DIRECTORY = abspath(dirname(__file__))
+HELPER_DIRECTORY = join(SCRIPT_DIRECTORY, "subdir")
+
 print("stdout output")
 def assert_path_contents(path, expected_contents):
     if not exists(path):
@@ -68,18 +76,27 @@ assert len(listdir(dirname(index_path))) == 2
 assert len(listdir(join(dirname(dirname(index_path)), "seq"))) == 1
 output4_index_path = open(sys.argv[11], 'w')
 metadata_dir = dirname(sys.argv[13])
+output5 = open(sys.argv[15], 'w')
+legacy_galaxy_json = int(sys.argv[16])
 output_metadata_path = join(metadata_dir, "metadata_output")
 try:
     assert_path_contents(sys.argv[2], "Hello world input!!@!")
     assert_path_contents(sys.argv[8], "INPUT_EXTRA_CONTENTS")
     assert_path_contents(sys.argv[13], "meta input")
+    assert_path_contents(sys.argv[14], "INPUT METADATA CONTENTS...")
     contents = config_input.read(1024)
     output.write(contents)
     open("workdir_output", "w").write("WORK DIR OUTPUT")
     open("env_test", "w").write(getenv("TEST_ENV", "DEFAULT"))
     open("rewrite_action_test", "w").write(sys.argv[12])
     output2.write(output2_contents)
-    with open("galaxy.json", "w") as f: f.write("GALAXY_JSON")
+    if legacy_galaxy_json:
+        with open("galaxy.json", "w") as f: f.write('''{"filename": "gjson_refer"}\n{"filename": "gjson_refer2"}''')
+    else:
+        with open("galaxy.json", "w") as f: f.write('''[{"filename": "gjson_refer"}, {"filename": "gjson_refer2"}]''')
+
+    with open("gjson_refer", "w") as f: f.write('''gjson_refer_contents''')
+    with open("gjson_refer2", "w") as f: f.write('''gjson_refer_contents2''')
     with open(output_metadata_path, "w") as f: f.write("meta output")
     output3.write(getenv("MOO", "moo_default"))
     output1_extras_path = "%s_files" % sys.argv[3][0:-len(".dat")]
@@ -87,6 +104,7 @@ try:
     open(join(output1_extras_path, "extra"), "w").write("EXTRA_OUTPUT_CONTENTS")
     version_output.write("1.0.1")
     output4_index_path.write(index_path)
+    output5.write(str(exists(join(HELPER_DIRECTORY, "helper.R"))))
 finally:
     output.close()
     config_input.close()
@@ -108,6 +126,7 @@ HELP_DISABLE_CLEANUP = ("Specify to disable cleanup after the job, this "
                         "is useful to checking the files generated during "
                         "the job and stored on the Pulsar server.")
 HELP_JOB_ID = "Submit the Pulsar job with this 'external' id."
+HELP_DEBUG = "Enable debug log output from Pulsar client"
 
 EXPECTED_OUTPUT = b"hello world output"
 EXAMPLE_UNICODE_TEXT = u'єχαмρℓє συтρυт'
@@ -125,10 +144,19 @@ class MockTool(object):
         self.tool_dir = tool_dir
 
 
+class TestRequiredFilesObject:
+    # it would be nice to use a RequiredFiles but that hasn't been published to PyPI yet, so just
+    # stick to the interface here.
+
+    def find_required_files(self, tool_directory):
+        return ["script.py", "subdir/helper.R"]
+
+
 def run(options):
+    logging.basicConfig(level=logging.DEBUG)
     waiter = None
     try:
-        temp_directory = tempfile.mkdtemp()
+        temp_directory = tempfile.mkdtemp(prefix='pulsar-check-client.')
         temp_index_dir = os.path.join(temp_directory, "idx", "bwa")
         temp_index_dir_sibbling = os.path.join(temp_directory, "idx", "seq")
         temp_shared_dir = os.path.join(temp_directory, "shared", "test1")
@@ -136,9 +164,11 @@ def run(options):
         temp_metadata_dir = os.path.join(temp_directory, "m")
         temp_false_working_dir = os.path.join(temp_metadata_dir, "working")
         temp_tool_dir = os.path.join(temp_directory, "t")
+        temp_tool_sub_dir = os.path.join(temp_tool_dir, "subdir")
 
         __makedirs([
             temp_tool_dir,
+            temp_tool_sub_dir,
             temp_work_dir,
             temp_index_dir,
             temp_index_dir_sibbling,
@@ -149,14 +179,17 @@ def run(options):
 
         temp_input_path = os.path.join(temp_directory, "dataset_0.dat")
         temp_input_extra_path = os.path.join(temp_directory, "dataset_0_files", "input_subdir", "extra")
+        temp_input_metadata_path = os.path.join(temp_directory, "metadata", "12312231231231.dat")
         temp_index_path = os.path.join(temp_index_dir, "human.fa")
 
         temp_config_path = os.path.join(temp_work_dir, "config.txt")
         temp_tool_path = os.path.join(temp_directory, "t", "script.py")
+        temp_tool_helper_path = os.path.join(temp_directory, "t", "subdir", "helper.R")
         temp_output_path = os.path.join(temp_directory, "dataset_1.dat")
         temp_output2_path = os.path.join(temp_directory, "dataset_2.dat")
         temp_output3_path = os.path.join(temp_directory, "dataset_3.dat")
         temp_output4_path = os.path.join(temp_directory, "dataset_4.dat")
+        temp_output5_path = os.path.join(temp_directory, "dataset_5.dat")
         temp_version_output_path = os.path.join(temp_directory, "GALAXY_VERSION_1234")
         temp_output_workdir_destination = os.path.join(temp_directory, "dataset_77.dat")
         temp_output_workdir = os.path.join(temp_work_dir, "env_test")
@@ -167,9 +200,11 @@ def run(options):
 
         __write_to_file(temp_input_path, b"Hello world input!!@!")
         __write_to_file(temp_input_extra_path, b"INPUT_EXTRA_CONTENTS")
+        __write_to_file(temp_input_metadata_path, b"INPUT METADATA CONTENTS...")
         __write_to_file(temp_config_path, EXPECTED_OUTPUT)
         __write_to_file(temp_metadata_path, "meta input")
         __write_to_file(temp_tool_path, TEST_SCRIPT)
+        __write_to_file(temp_tool_helper_path, b"helper R file")
         __write_to_file(temp_index_path, b"AGTC")
         # Implicit files that should also get transferred since depth > 0
         __write_to_file("%s.fai" % temp_index_path, b"AGTC")
@@ -178,6 +213,7 @@ def run(options):
         empty_input = u"/foo/bar/x"
 
         test_unicode = getattr(options, "test_unicode", False)  # TODO Switch this in integration tests
+        legacy_galaxy_json = getattr(options, "legacy_galaxy_json", False)
         cmd_text = EXAMPLE_UNICODE_TEXT if test_unicode else "Hello World"
         command_line_params = (
             temp_tool_path,
@@ -194,21 +230,34 @@ def run(options):
             temp_output4_path,
             temp_shared_dir,
             temp_metadata_path,
+            temp_input_metadata_path,
+            temp_output5_path,
+            "1" if legacy_galaxy_json else "0",
         )
         assert os.path.exists(temp_index_path)
-        command_line = u'python %s "%s" "%s" "%s" "%s" "%s" "%s" "%s" "%s" "%s" "%s" "%s" "%s" "%s"' % command_line_params
+        command_line = u'python %s "%s" "%s" "%s" "%s" "%s" "%s" "%s" "%s" "%s" "%s" "%s" "%s" "%s" "%s" "%s" "%s"' % command_line_params
         config_files = [temp_config_path]
-        input_files = [temp_input_path, empty_input]
+        client_inputs = []
+        client_inputs.append(ClientInput(temp_input_path, CLIENT_INPUT_PATH_TYPES.INPUT_PATH))
+        client_inputs.append(ClientInput(temp_input_path, CLIENT_INPUT_PATH_TYPES.INPUT_PATH))
+        # Reverting empty input handling added in:
+        #  https://github.com/galaxyproject/pulsar/commit/2fb36ba979cf047a595c53cdef833cae79cbb380
+        # Seems like it really should cause a failure.
+        # client_inputs.append(ClientInput(empty_input, CLIENT_INPUT_PATH_TYPES.INPUT_PATH))
+        client_inputs.append(ClientInput(os.path.join(temp_directory, "dataset_0_files"), CLIENT_INPUT_PATH_TYPES.INPUT_EXTRA_FILES_PATH))
+        client_inputs.append(ClientInput(temp_input_metadata_path, CLIENT_INPUT_PATH_TYPES.INPUT_METADATA_PATH))
         output_files = [
             temp_output_path,
             temp_output2_path,
             temp_output3_path,
             temp_output4_path,
+            temp_output5_path,
             temp_output_workdir_destination,
-            temp_output_workdir_destination2
+            temp_output_workdir_destination2,
         ]
         client, client_manager = __client(temp_directory, options)
         waiter = Waiter(client, client_manager)
+        galaxy_json_type = "legacy_galaxy" if legacy_galaxy_json else "galaxy"
         client_outputs = ClientOutputs(
             working_directory=temp_work_dir,
             metadata_directory=temp_metadata_dir,
@@ -218,13 +267,13 @@ def run(options):
             ],
             output_files=output_files,
             version_file=temp_version_output_path,
+            dynamic_file_sources=[{"type": galaxy_json_type, "path": "galaxy.json"}],
         )
-
         job_description = ClientJobDescription(
             command_line=command_line,
             tool=MockTool(temp_tool_dir),
             config_files=config_files,
-            input_files=input_files,
+            client_inputs=ClientInputs(client_inputs),
             client_outputs=client_outputs,
             working_directory=temp_work_dir,
             metadata_directory=temp_metadata_dir,
@@ -241,7 +290,11 @@ def run(options):
         __finish(options, client, client_outputs, result_status)
         __assert_contents(temp_output_path, EXPECTED_OUTPUT, result_status)
         __assert_contents(temp_output2_path, cmd_text, result_status)
-        __assert_contents(os.path.join(temp_work_dir, "galaxy.json"), b"GALAXY_JSON", result_status)
+        if not legacy_galaxy_json:
+            expected_contents = b'''[{"filename": "gjson_refer"}, {"filename": "gjson_refer2"}]'''
+            __assert_contents(os.path.join(temp_work_dir, "galaxy.json"), expected_contents, result_status)
+        __assert_contents(os.path.join(temp_work_dir, "gjson_refer"), b"gjson_refer_contents", result_status)
+        __assert_contents(os.path.join(temp_work_dir, "gjson_refer2"), b"gjson_refer_contents2", result_status)
         __assert_contents(os.path.join(temp_directory, "dataset_1_files", "extra"), b"EXTRA_OUTPUT_CONTENTS", result_status)
         __assert_contents(os.path.join(temp_metadata_dir, "metadata_output"), b"meta output", result_status)
         if getattr(options, "test_rewrite_action", False):
@@ -254,6 +307,8 @@ def run(options):
         else:
             __assert_contents(temp_output3_path, "moo_default", result_status)
         __assert_has_rewritten_bwa_path(client, temp_output4_path)
+        if getattr(options, "explicit_tool_declarations", False):
+            __assert_contents(temp_output5_path, str(True), result_status)
         __exercise_errors(options, client, temp_output_path, temp_directory)
         client_manager.shutdown()
     except BaseException:
@@ -263,7 +318,8 @@ def run(options):
     finally:
         if waiter is not None:
             waiter.shutdown()
-        shutil.rmtree(temp_directory)
+        if getattr(options, 'cleanup', True):
+            shutil.rmtree(temp_directory)
 
 
 class Waiter(object):
@@ -332,8 +388,8 @@ def __assert_has_rewritten_bwa_path(client, temp_output4_path):
     if client.default_file_action != "none":
         rewritten_index_path = open(temp_output4_path, 'r', encoding='utf-8').read()
         # Path written to this file will differ between Windows and Linux.
-        if re.search(r"123456[/\\]unstructured[/\\]\w+[/\\]bwa[/\\]human.fa", rewritten_index_path) is None:
-            raise AssertionError("[%s] does not container rewritten path." % rewritten_index_path)
+        if re.search(r"%s[/\\]unstructured[/\\]\w+[/\\]bwa[/\\]human.fa" % client.job_id, rewritten_index_path) is None:
+            raise AssertionError("[%s] does not contain rewritten path." % rewritten_index_path)
 
 
 def __exercise_errors(options, client, temp_output_path, temp_directory):
@@ -399,6 +455,10 @@ def extract_client_options(options):
         client_options["jobs_directory"] = getattr(options, "jobs_directory")
     if hasattr(options, "files_endpoint"):
         client_options["files_endpoint"] = getattr(options, "files_endpoint")
+    if hasattr(options, "k8s_enabled"):
+        client_options["k8s_enabled"] = getattr(options, "k8s_enabled")
+    if hasattr(options, "container"):
+        client_options["container"] = getattr(options, "container")
     return client_options
 
 
@@ -452,7 +512,12 @@ def __extra_job_description_kwargs(options):
     env = []
     if test_env:
         env.append(dict(name="TEST_ENV", value="TEST_ENV_VALUE"))
-    return dict(dependencies_description=dependencies_description, env=env)
+    container = getattr(options, "container", None)
+    remote_pulsar_app_config = getattr(options, "remote_pulsar_app_config", None)
+    rval = dict(dependencies_description=dependencies_description, env=env, container=container, remote_pulsar_app_config=remote_pulsar_app_config)
+    if getattr(options, "explicit_tool_declarations", False):
+        rval["tool_directory_required_files"] = TestRequiredFilesObject()
+    return rval
 
 
 def __finish(options, client, client_outputs, result_status):
@@ -486,6 +551,9 @@ def main(argv=None):
     parser.add_option('--suppress_output', default=False, action="store_true", help=HELP_SUPPRESS_OUTPUT)
     parser.add_option('--disable_cleanup', dest="cleanup", default=True, action="store_false", help=HELP_DISABLE_CLEANUP)
     parser.add_option('--job_id', default="123456", help=HELP_JOB_ID)
+    parser.add_option('--explicit_tool_declarations', default=False, action="store_true")
+    parser.add_option('--legacy_galaxy_json', default=False, action="store_true")
+    parser.add_option('--debug', default=False, action="store_true", help=HELP_DEBUG)
     (options, args) = parser.parse_args(argv)
     run(options)
 
